# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import xprec


def _compare_ufunc(ufunc, *args, ulps=1):
    fx_d = ufunc(*args)
    fx_q = ufunc(*(a.astype(xprec.ddouble) for a in args)).astype(float)

    # Ensure relative accuracy of 2 ulps
    np.testing.assert_array_almost_equal_nulp(fx_d, fx_q, ulps)


def test_log():
    x = np.geomspace(1e-300, 1e300, 1953)
    _compare_ufunc(np.log, x)

    zeroq = xprec.ddouble.type(0)
    assert np.isinf(np.log(zeroq))


def test_sqrt():
    x = np.geomspace(1e-300, 1e300, 1953)
    _compare_ufunc(np.sqrt, x)


def test_exp():
    x = np.geomspace(1e-300, 700, 4953)
    x = np.hstack([-x[::-1], 0, x])
    _compare_ufunc(np.exp, x)

    # Unfortunately, on Windows expm1 is less precise, so we need to increase
    # the tolerance slightly
    _compare_ufunc(np.expm1, x, ulps=2)


def test_cosh():
    x = np.geomspace(1e-300, 700, 4953)
    x = np.hstack([-x[::-1], 0, x])
    _compare_ufunc(np.cosh, x)
    _compare_ufunc(np.sinh, x)

    thousand = xprec.ddouble.type(1000)
    assert np.isinf(np.cosh(thousand))
    assert np.isinf(np.cosh(-thousand))


def test_hypot():
    x = np.geomspace(1e-300, 1e260, 47)
    x = np.hstack([-x[::-1], 0, x])
    _compare_ufunc(np.hypot, x[:,None], x[None,:])


def test_modf():
    ulps = 1
    x = np.linspace(-100, 100, 100)
    x_d = x.astype(xprec.ddouble)

    fx_d = np.modf(x)
    fx_q = np.modf(x_d)

    # Ensure relative accuracy of 1 ulp
    np.testing.assert_array_almost_equal_nulp(fx_d[0], fx_q[0].astype(float), ulps)
    np.testing.assert_array_almost_equal_nulp(fx_d[1], fx_q[1].astype(float), ulps)


def test_power():
    x = np.linspace(0, 100, 100)
    _compare_ufunc(np.power, x[:,None], x[None,:])


def test_arctan2():
    x = np.linspace(-100, 100, 100)
    _compare_ufunc(np.arctan2, x[:,None], x[None,:], ulps=2)


def test_arcsin():
    x = np.linspace(-1, 1, 100)
    _compare_ufunc(np.arcsin, x, ulps=2)


def test_arccos():
    x = np.linspace(-1, 1, 100)
    _compare_ufunc(np.arccos, x)


def test_arctan():
    x = np.linspace(-100, 100, 100)
    _compare_ufunc(np.arctan, x)


def test_arccosh():
    x = np.linspace(1, 100, 100)
    _compare_ufunc(np.arccosh, x)


def test_arcsinh():
    x = np.linspace(-100, 100, 100)
    _compare_ufunc(np.arcsinh, x)


def test_arctanh():
    x = np.linspace(-0.99, 0.99, 100)
    _compare_ufunc(np.arctanh, x)



# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
#
# Some of the code in this module is adapted from the LAPACK reference
# implementation.
import numpy as np
from warnings import warn

from . import ddouble
from . import _dd_linalg

norm = _dd_linalg.norm
givens = _dd_linalg.givens
householder = _dd_linalg.householder
rank1update = _dd_linalg.rank1update


def qr(A, reflectors=False):
    """QR decomposition without pivoting.

    Decomposes a `(m, n)` matrix `A` into the product:

        A == Q @ R

    where `Q` is an `(m, m)` orthogonal matrix and `R` is a `(m, n)` upper
    triangular matrix.  No pivoting is used.
    """
    R = np.array(A)
    m, n = R.shape
    k = min(m, n)

    Q = np.zeros((k, m), A.dtype)
    for i in range(k):
        householder_update(R[i:,i:], Q[i:,i:])
    if not reflectors:
        I = np.eye(m, dtype=A.dtype)
        Q = householder_apply(Q, I)
    return Q, R


def rrqr(A, tol=5e-32, reflectors=False):
    """Truncated rank-revealing QR decomposition with full column pivoting.

    Decomposes a `(m, n)` matrix `A` into the product:

        A[:,piv] == Q @ R

    where `Q` is an `(m, k)` isometric matrix, `R` is a `(k, n)` upper
    triangular matrix, `piv` is a permutation vector, and `k` is chosen such
    that the relative tolerance `tol` is met in the equality above.
    """
    R = np.array(A)
    m, n = R.shape
    k = min(m, n)

    Q = np.zeros((m, k), A.dtype)
    jpvt = np.arange(n)
    norms = norm(A.T)
    xnorms = norms.copy()
    TOL3Z = np.finfo(float).eps
    for i in range(k):
        pvt = i + np.argmax(norms[i:])
        if i != pvt:
            R[:,[i, pvt]] = R[:,[pvt, i]]
            jpvt[[i, pvt]] = jpvt[[pvt, i]]
            norms[pvt] = norms[i]
            xnorms[pvt] = xnorms[i]

        householder_update(R[i:,i:], Q[i:,i:])

        js = (i + 1) + norms[i + 1:].nonzero()[0]
        temp = np.abs(R[i,js]) / norms[js]
        temp = np.fmax(0.0, (1 + temp)*(1 - temp))
        temp2 = temp * np.square(norms[js] / xnorms[js])

        wheresmall = temp2 < TOL3Z
        jsmall = js[wheresmall]
        upd_norms = norm(R[i+1:,jsmall].T)
        norms[jsmall] = upd_norms
        xnorms[jsmall] = upd_norms
        jbig = js[~wheresmall]
        norms[jbig] *= np.sqrt(temp[~wheresmall])

        if tol is not None:
            acc = np.abs(R[i,i] / R[0,0])
            if acc < tol:
                k = i + 1
                Q = Q[:,:k]
                R = R[:k,:]
                break

    if not reflectors:
        I = np.eye(m, k, dtype=A.dtype)
        Q = householder_apply(Q, I)
    return Q, R, jpvt


def svd(A, full_matrices=True):
    """Truncated singular value decomposition.

    Decomposes a `(m, n)` matrix `A` into the product:

        A == U @ (s[:,None] * VT)

    where `U` is a `(m, k)` matrix with orthogonal columns, `VT` is a `(k, n)`
    matrix with orthogonal rows and `s` are the singular values, a set of `k`
    nonnegative numbers in non-ascending order and `k = min(m, n)`.
    """
    A = np.asarray(A)
    m, n = A.shape
    if m < n:
        U, s, VT = svd(A.T, full_matrices)
        return VT.T, s, U.T

    Q, B, RT = bidiag(A)
    for _ in range(20 * n):
        if svd_bidiag_step(Q, B, RT):
            break
    else:
        warn("Did not converge")

    U, s, VH = svd_normalize(Q, B.diagonal(), RT)
    if not full_matrices:
        U = U[:,:n]
    return U, s, VH


def svd_trunc(A, tol=5e-32, method='jacobi', max_iter=20):
    """Truncated singular value decomposition.

    Decomposes a `(m, n)` matrix `A` into the product:

        A == U @ (s[:,None] * VT)

    where `U` is a `(m, k)` matrix with orthogonal columns, `VT` is a `(k, n)`
    matrix with orthogonal rows and `s` are the singular values, a set of `k`
    nonnegative numbers in non-ascending order.  The SVD is truncated in the
    sense that singular values below `tol` are discarded.
    """
    # RRQR is an excellent preconditioner for Jacobi.  One should then perform
    # Jacobi on RT
    Q, R, p = rrqr(A, tol)
    if method == 'jacobi':
        U, s, VT = svd_jacobi(R.T, tol, max_iter)
    elif method == 'golub-kahan':
        U, s, VT = svd(R.T, full_matrices=False)
    else:
        raise ValueError("invalid method")

    # Reconstruct A from QRs
    U_A = Q @ VT.T
    VT_B = U.T[:, p.argsort()]
    return U_A, s, VT_B


def bidiag(A, reflectors=False, force_structure=False):
    """Biadiagonalizes an arbitray rectangular matrix.

    Decomposes a `(m, n)` matrix `A` into the product:

        A == Q @ B @ RT

    where `Q` is a `(m, m)` orthogonal matrix, `RT` is a `(n, n)` orthogonal
    matrix, and `B` is a bidiagonal matrix, where the upper diagonal is
    nonzero for `m >= n` and the lower diagonal is nonzero for `m < n`.
    """
    A = np.asarray(A)
    m, n = A.shape
    if m < n:
        Q, B, RT = bidiag(A.T, reflectors)
        return RT.T, B.T, Q.T

    rq = n - (m == n)
    B = A.copy()
    Q = np.zeros_like(B)
    R = np.zeros_like(B[:n,:n])

    for j in range(n-2):
        householder_update(B[j:,j:], Q[j:,j:])
        householder_update(B[j:,j+1:].T, R[j+1:,j+1:])
    for j in range(n-2, rq):
        householder_update(B[j:,j:], Q[j:,j:])

    if force_structure:
        d = B.diagonal().copy()
        e = B.diagonal(1).copy()
        B[...] = 0
        i = np.arange(n)
        B[i, i] = d
        B[i[:-1], i[:-1]+1] = e
    if not reflectors:
        Q = householder_apply(Q, np.eye(m, dtype=B.dtype))
        R = householder_apply(R, np.eye(n, dtype=B.dtype))
    return Q, B, R.T


def svd_jacobi(A, tol=5e-32, max_iter=20):
    """Singular value decomposition using Jacobi rotations."""
    U = A.copy()
    m, n = U.shape
    if m < n:
        raise RuntimeError("expecting tall matrix")

    VT = np.eye(n, dtype=U.dtype)
    offd = np.empty((), ddouble)

    limit = tol * np.linalg.norm(U[:n,:n], 'fro')
    for _ in range(max_iter):
        _dd_linalg.jacobi_sweep(U, VT, out=(U, VT, offd))
        if offd <= limit:
            break
    else:
        warn("Did not converge")

    s = norm(U.T)
    U = U / s
    return U, s, VT


def householder_update(A, Q):
    """Reflects the zeroth column onto a multiple of the unit vector"""
    beta, v = householder(A[:,0])
    w = -beta * (A.T @ v)
    rank1update(A, v, w, out=A)
    Q[0,0] = beta
    Q[1:,0] = v[1:]


def householder_apply(H, Q):
    """Applies a set of reflectors to a matrix"""
    H = np.asarray(H)
    Q = Q.copy()
    m, r = H.shape
    if Q.shape[0] != m:
        raise ValueError("invalid shape")
    if Q.shape[1] < r:
        raise ValueError("invalid shape")
    for j in range(r-1, -1, -1):
        beta = H[j,j]
        if np.equal(beta, 0):
            continue
        v = np.empty_like(H[j:,0])
        v[0] = 1
        v[1:] = H[j+1:,j]
        Qpart = Q[j:,j:]
        w = -beta * (Qpart.T @ v)
        rank1update(Qpart, v, w, out=Qpart)
    return Q


def svd_normalize(U, d, VH):
    """Given a SVD-like decomposition, normalize"""
    # Invert
    n = d.size
    VH[np.signbit(d)] = -VH[np.signbit(d)]
    d = np.abs(d)

    # Sort
    order = np.argsort(d)[::-1]
    d = d[order]
    VH = VH[order]
    U = U.copy()
    U[:,:n] = U[:,order]
    return U, d, VH


def svd_bidiag_step(Q, B, RT):
    """Single SVD step for a bidiagonal matrix"""
    d = B.diagonal().copy()
    e = np.hstack([B.diagonal(1), 0.0])

    p, q = bidiag_partition(d, e)
    if q <= 1:
        return True

    d_part = d[p:q]
    e_part = e[p:q]
    rot = np.empty((d_part.size, 4), d.dtype)
    _dd_linalg.golub_kahan_chase(d_part, e_part, out=(d_part, e_part, rot))

    i = np.arange(p, q)
    B[i, i] = d_part
    B[i[:-1], i[:-1]+1] = e_part[:-1]

    rot_Q = rot[:, 2:]
    rot_R = rot[:, :2]
    QT_part = Q[:, p:q].T
    RT_part = RT[p:q, :]
    _dd_linalg.givens_seq(rot_Q, QT_part, out=QT_part)
    _dd_linalg.givens_seq(rot_R, RT_part, out=RT_part)
    return False


def bidiag_partition(d, e, eps=5e-32):
    """Partition bidiagonal matrix into blocks for implicit QR.

    Return `p,q` which partions a bidiagonal `B` matrix into three blocks:

      - B[0:p, 0:p], an arbitrary bidiaonal matrix
      - B[p:q, p:q], a matrix with all off-diagonal elements nonzero
      - B[q:,  q:],  a diagonal matrix
    """
    abs_e = np.abs(e)
    abs_d = np.abs(d)
    e_zero = abs_e <= eps * (abs_d + abs_e)
    e[e_zero] = 0

    q = _find_last(~e_zero) + 1
    if q <= 0:
        return 0, 0
    p = _find_last(e_zero[:q]) + 1
    return p, q + 1


def _find_last(a, axis=-1):
    a = a.astype(bool)
    maxloc = a.shape[axis] - 1 - a[::-1].argmax(axis)
    return np.where(a[maxloc], maxloc, -1)



import os

HEREPATH = os.path.abspath(os.path.dirname(__file__))
print("HEREPATH", HEREPATH)
ROOTDIR = os.path.abspath(os.path.join(HEREPATH, os.path.pardir))
PYSRCDIR = os.path.join(ROOTDIR, "pysrc", "xprec")
CSCRDIR = os.path.join(ROOTDIR, "csrc")

def check_whitespace(files):
    errors = []
    blank = 0
    lineno = 0
    line = ""
    def add_error(fmt, *params):
        errors.append((fname, lineno, line, fmt.format(*params)))

    for fname in files:
        with open(fname, "r") as file:
            line = ""
            for lineno, line in enumerate(file, start=1):
                if line[-1:] != '\n':
                    add_error("file must end in blank line")
                line = line[:-1]
                if line:
                    blank = 0
                else:
                    blank += 1
                if line[-1:] == '\r':
                    add_error("file must only have unix line endings")
                if line[-1:] == ' ':
                    add_error("line ends in whitespace")
                if '\t' in line:
                    add_error("line contains tab characters")
                if len(line) > 90:
                    add_error("line is too long: {:d} chars", len(line))
            # end of file
            if blank != 0:
                add_error("file has {:d} superflouos blank lines", blank)

    msg = ""
    for fname, lineno, line, lmsg in errors:
        msg += "{}:{}: {}\n".format(fname.name, lineno, lmsg)
    if msg:
        raise ValueError("Whitespace errors\n" + msg)


def all_files(path, ext):
    for entry in os.scandir(path):
        if entry.is_file() and entry.name.endswith(ext):
            yield entry


def test_ws_testdir():
    check_whitespace(all_files(HEREPATH, ".py"))


def test_ws_setup():
    check_whitespace(all_files(ROOTDIR, ".py"))


def test_ws_pysrcdir():
    check_whitespace(all_files(PYSRCDIR, ".py"))


def test_ws_csrcdir():
    check_whitespace(all_files(CSCRDIR, ".c"))
    check_whitespace(all_files(CSCRDIR, ".h"))



# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import pytest

import xprec

EPS = xprec.finfo(xprec.ddouble).eps

try:
    import mpmath
except ImportError:
    pytest.skip("No mpmath library avaiable", allow_module_level=True)
else:
    mpmath.mp.prec = 120


def mpf_for_xprec(x):
    """Converts xprec.ddouble array to array of mpmath mpf scalars"""
    x = np.asarray(x)
    if x.dtype != xprec.ddouble:
        raise ValueError("dtype shall be ddouble")

    x_flat = x.ravel()
    x_hi = x_flat.astype(float)
    x_lo = (x_flat - x_hi).astype(float)
    x_mpf = np.array(list(map(mpmath.mpf, x_hi)))
    x_mpf += x_lo
    return x_mpf.reshape(x.shape)


def map_mpmath(fn, x):
    x = np.asarray(x)
    x_flat = x.ravel()
    y_flat = np.array(list(map(fn, x_flat)), dtype=object)
    y = y_flat.reshape(x.shape)
    return y


def check_unary(mpmath_fn, numpy_fn, x, rtol):
    y_ref = map_mpmath(mpmath_fn, x)
    y_our = numpy_fn(x.astype(xprec.ddouble))
    y_float = y_ref.astype(float)

    diff = (y_ref - mpf_for_xprec(y_our)).astype(float)
    ok = np.abs(diff) <= rtol * np.abs(y_float)
    if not ok.all():
        x = x[~ok]
        y_float = y_float[~ok]
        y_our = y_our[~ok]
        diff = diff[~ok]
        reldiff = diff / np.abs(y_float)

        msg = f"{'x':>13s} {'mpmath':>13s} {'xprec':>13s} {'rel diff':>13s}\n"
        msg += "\n".join(f"{xi:13g} {y_refi:13g} {y_ouri:13g} {reldiffi:13g}"
                         for xi, y_refi, y_ouri, reldiffi, _
                         in zip(x, y_float, y_our, reldiff, range(10))
                         )
        raise ValueError(f"not equal to rtol = {rtol:3g}\n" + msg)



def test_sqrt():
    # Once the low part of the ddouble becomes a denormal number, we
    # are in trouble, so we truncate the lower end of the range by
    # another 16 digits
    x = np.geomspace(1e-292, 1e307, 1953)
    check_unary(mpmath.sqrt, np.sqrt, x, 2*EPS)


def test_log():
    x = np.reciprocal(np.geomspace(1e-292, 1e307, 1953))
    check_unary(mpmath.log, np.log, x, 70 * EPS)


def test_exp():
    x = np.geomspace(1e-280, 670, 1511)
    x = np.hstack([-x[::-1], 0, x])
    check_unary(mpmath.exp, np.exp, x, 60 * EPS)
    check_unary(mpmath.expm1, np.expm1, x, 60 * EPS)

    check_unary(mpmath.sinh, np.sinh, x, 60 * EPS)
    check_unary(mpmath.cosh, np.cosh, x, 60 * EPS)
    check_unary(mpmath.tanh, np.tanh, x, 60 * EPS)


def test_sincos():
    x = np.geomspace(1e-280, 4.8 * np.pi, 1511)
    x = np.hstack([-x[::-1], 0, x])
    check_unary(mpmath.sin, np.sin, x, 2 * EPS)
    check_unary(mpmath.cos, np.cos, x, 2 * EPS)



# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
"""
Extension module for numpy providing the `ddouble` data type.

Loading this module registers an additional scalar data type `ddouble` with
numpy implementing double-double arithmetic.  You can use use the data type
by passing `dtype=xprec.ddouble` to numpy functions.

Example:

    import numpy as np
    from xprec import ddouble

    x = np.arange(5, dtype=ddouble)
    print(2 * x)

"""
__version__ = "1.4.1"

import numpy as _np

from . import _dd_ufunc
from . import _dd_linalg    # needed for matmul

ddouble = _dd_ufunc.dtype


def finfo(dtype):
    dtype = _np.dtype(dtype)
    try:
        finfo_dunder = dtype.type.__finfo__
    except AttributeError:
        return _np.finfo(dtype)
    else:
        return finfo_dunder()



# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np
import pytest

from xprec import ddouble


COMPATIBLE_DTYPES = [
    np.int8, np.int16, np.int32, np.int64, np.bool_, np.float32, np.float64,
    np.uint8, np.uint16, np.uint32, np.uint64,
    ]


@pytest.mark.parametrize('other', COMPATIBLE_DTYPES)
def test_cast_from(other):
    assert np.can_cast(other, ddouble, 'unsafe')
    assert np.can_cast(other, ddouble, 'safe')

    x = np.eye(3, dtype=other)
    y = x.astype(ddouble)
    assert (x == y).all()


@pytest.mark.parametrize('other', COMPATIBLE_DTYPES)
def test_cast_to(other):
    assert np.can_cast(ddouble, other, 'unsafe')
    assert not np.can_cast(ddouble, other, 'safe')

    x = np.eye(3, dtype=ddouble)
    y = x.astype(other)
    assert (x == y).all()


def test_i64():
    x = np.int64((1 << 62) + 1)
    assert x == x.astype(ddouble).astype(x.dtype)

    y = -x
    assert (y + 1) == (y.astype(ddouble) + 1).astype(x.dtype)

    x = x.astype(np.uint64)
    assert x == x.astype(ddouble).astype(x.dtype)



# Setup script - embracing the setuptools madness.
#
# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import io
import os.path
import platform
import re

from setuptools import setup, find_packages
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext as BuildExt


def readfile(*parts):
    """Return contents of file with path relative to script directory"""
    herepath = os.path.abspath(os.path.dirname(__file__))
    fullpath = os.path.join(herepath, *parts)
    with io.open(fullpath, 'r') as f:
        return f.read()


def extract_version(*parts):
    """Extract value of __version__ variable by parsing python script"""
    initfile = readfile(*parts)
    version_re = re.compile(r"(?m)^__version__\s*=\s*['\"]([^'\"]*)['\"]")
    match = version_re.search(initfile)
    return match.group(1)


def rebase_links(text, base_url):
    """Rebase links to doc/ directory to ensure they work online."""
    doclink_re = re.compile(
                        r"(?m)^\s*\[\s*([^\]\n\r]+)\s*\]:\s*(doc/[./\w]+)\s*$")
    result, nsub = doclink_re.subn(r"[\1]: %s/\2" % base_url, text)
    return result


def append_if_absent(list, arg):
    """Append argument to list if absent"""
    if arg not in list:
        list.append(arg)


def get_flags_dict(exec):
    # First, let us clean up the mess of compiler options a little bit:  Move
    # flags out into a dictionary, thereby removing the myriad of duplicates
    cc_so, *cflags_so_list = exec
    cflags_curr = None
    cflags_so = {}
    for arg in cflags_so_list:
        if arg.startswith("-"):
            if cflags_curr is not None:
                cflags_so[cflags_curr] = None
                cflags_curr = None
            arg = arg.split("=", 1)
            if len(arg) == 1:
                cflags_curr, = arg
            else:
                k, v = arg
                cflags_so[k] = v
        else:
            if cflags_curr is None:
                raise ValueError("expected flag" + str(exec))
            cflags_so[cflags_curr] = arg
            cflags_curr = None
    if cflags_curr is not None:
        cflags_so[cflags_curr] = None

    return cc_so, cflags_so


def make_exec_string(cc_so, cflags_so):
    # Now update the flags
    cflags_so = [k + ("=" + v if v is not None else "")
                 for (k,v) in cflags_so.items()]
    return [cc_so] + cflags_so


class OptionsMixin:
    _convert_to_bool = {"true": True, "false": False}
    user_options = [
        ("with-openmp=", None, "use openmp to build (default: false)"),
        ("opt-arch=", None, "optimized for architecture"),
        ("numpy-include-dir=", None, "numpy include directory"),
        ]

    def initialize_options(self):
        super().initialize_options()
        self.with_openmp = None
        self.numpy_include_dir = None
        self.opt_arch = None

    def finalize_options(self):
        if self.with_openmp is not None:
            self.with_openmp = self._convert_to_bool[self.with_openmp.lower()]
        if self.opt_arch is not None:
            self.opt_arch = self._convert_to_bool[self.opt_arch.lower()]
        if self.numpy_include_dir is not None:
            if not os.path.isdir(self.numpy_include_dir):
                raise ValueError("include directory must exist")
        super().finalize_options()


class BuildExtWithNumpy(OptionsMixin, BuildExt):
    """Wrapper class for building numpy extensions"""
    user_options = BuildExt.user_options + OptionsMixin.user_options

    def build_extensions(self):
        """Modify paths according to options"""
        # This must be deferred to build time, because that is when
        # self.compiler starts being a compiler instance (before, it is
        # a flag)  *slow-clap*
        # compiler type is either 'unix', 'msvc' or 'mingw'
        compiler_type = self.compiler.compiler_type

        compiler_binary = getattr(self.compiler, 'compiler', [''])[0]
        compiler_binary = os.path.basename(compiler_binary)
        compiler_make = ''
        if 'gcc' in compiler_binary or 'g++' in compiler_binary:
            compiler_make = 'gcc'
        elif 'clang' in compiler_binary:
            compiler_make = 'clang'
        elif 'icc' in compiler_binary:
            compiler_make = 'icc'
        elif compiler_type == 'msvc':
            # See msvccompiler.py:206 - a comment worth reading in its
            # entirety.  distutils sets up an abstraction which it immediately
            # break with its own derived classes.  *slow-clap*
            compiler_make = 'msvc'

        if compiler_type != 'msvc':
            new_flags = {"-Wextra": None, "-std": "c11"}
            # By default, we do not optimize for the architecture by default,
            # because this is harmful when building a binary package.
            if self.opt_arch:
                new_flags["-mtune"] = new_flags["-march"] = "native"

            cc_so, flags_dict = get_flags_dict(self.compiler.compiler_so)

            # Replace arch with march
            if "-arch" in flags_dict:
                flags_dict["-march"] = flags_dict.pop("-arch")

            flags_dict.update(new_flags)
            self.compiler.compiler_so = make_exec_string(cc_so, flags_dict)

        # This has to be set to false because MacOS does not ship openmp
        # by default.
        if self.with_openmp is None:
            self.with_openmp = platform.system() == 'Linux'

        # Numpy headers: numpy must be imported here rather than
        # globally, because otherwise it may not be available at the time
        # when the setup script is run.  *slow-cl ... ah, f*ck it.
        if self.numpy_include_dir is None:
            import numpy
            self.numpy_include_dir = numpy.get_include()

        for ext in self.extensions:
            append_if_absent(ext.include_dirs, self.numpy_include_dir)
            if self.with_openmp:
                append_if_absent(ext.extra_compile_args, '-fopenmp')
                append_if_absent(ext.extra_link_args, '-fopenmp')
                if compiler_make == 'clang':
                    append_if_absent(ext.extra_link_args, '-lomp')

        super().build_extensions()

    def get_source_files(self):
        """Return list of files to include in source dist"""
        # Specifying include_dirs= argument in Extension adds headers from that
        # directory to the sdist ... on some machines.  On others, not.  Note
        # that overriding sdist will not save you, since this is not called
        # from sdist.add_defaults(), as you might expect. (With setuptools, it
        # is never what you expect.)  Instead, sdist requires egg_info, which
        # hooks into a hidden manifest_maker class derived from sdist, where
        # add_defaults() called, the list passed back to sdist, sidestepping
        # the method in the orginal class.  Kudos.
        #
        # Really, if you have monkeys type out 1000 pages on typewriters, use
        # the result as toilet paper for a month, unfold it, scan it at 20 dpi,
        # and run it through text recognition software, it would still yield
        # better code than setuptools.
        source_files = super().get_source_files()
        header_regex = re.compile(r"\.(?:h|hh|hpp|hxx|H|HH|HPP|HXX)$")

        include_dirs = set()
        for ext in self.extensions:
            include_dirs.update(ext.include_dirs)
        for dir in include_dirs:
            for entry in os.scandir(dir):
                if not entry.is_file():
                    continue
                if not header_regex.search(entry.name):
                    continue
                source_files.append(entry.path)

        return source_files


VERSION = extract_version('pysrc', 'xprec', '__init__.py')
REPO_URL = "https://github.com/tuwien-cms/xprec"
DOCTREE_URL = "%s/tree/v%s" % (REPO_URL, VERSION)
LONG_DESCRIPTION = rebase_links(readfile('README.md'), DOCTREE_URL)

setup(
    name='xprec',
    version=VERSION,

    description='xprec precision numpy extension',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=' '.join([
        'double-double'
        ]),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        ],

    url=REPO_URL,
    author=', '.join([
        'Markus Wallerberger'
        ]),
    author_email='markus.wallerberger@tuwien.ac.at',

    python_requires='>=3',
    install_requires=[
        # we need matmul to be an ufunc -> 1.16
        'numpy>=1.16',
        ],
    extras_require={
        'test': ['pytest', 'mpmath'],
        },

    ext_modules=[
        Extension("xprec._dd_ufunc",
                  ["csrc/_dd_ufunc.c", "csrc/dd_arith.c"],
                  include_dirs=["csrc"]),
        Extension("xprec._dd_linalg",
                  ["csrc/_dd_linalg.c", "csrc/dd_arith.c", "csrc/dd_linalg.c"],
                  include_dirs=["csrc"]),
        ],
    setup_requires=[
        'numpy>=1.16'
        ],
    cmdclass={
        'build_ext': BuildExtWithNumpy
        },

    package_dir={'': 'pysrc'},
    packages=find_packages(where='pysrc'),
    )



# Copyright (C) 2021 Markus Wallerberger and others
# SPDX-License-Identifier: MIT
import numpy as np

import xprec
import xprec.linalg
from xprec import ddouble


def test_householder_vec():
    rng = np.random.RandomState(4711)
    xd = rng.random_sample(20)
    xq = np.array(xd, dtype=ddouble)

    betaq, vq = xprec.linalg.householder(xq)
    eq = xq - betaq * vq * (vq @ xq)
    np.testing.assert_allclose(eq[1:].astype(float), 0, atol=1e-31)


def test_bidiag():
    rng = np.random.RandomState(4711)
    m, n = 7, 5
    A = rng.normal(size=(m,n)).astype(ddouble)
    Q, B, RT = xprec.linalg.bidiag(A)
    diff = Q @ B @ RT - A

    # FIXME: too large precision goals
    np.testing.assert_allclose(diff.astype(float), 0, atol=1e-29)


def test_svd():
    rng = np.random.RandomState(4711)
    A = rng.randn(100, 84)

    U, s, VT = xprec.linalg.svd(A.astype(xprec.ddouble), full_matrices=False)
    R = U * s @ VT - A
    np.testing.assert_allclose(R.astype(float), 0, atol=5e-29, rtol=0)

    _, sx, _ = np.linalg.svd(A.astype(float), full_matrices=False)
    np.testing.assert_allclose(s, sx, atol=1e-14 * sx[0], rtol=0)


def test_givens():
    f, g = np.array([3.0, -2.0], dtype=ddouble)
    c, s, r = xprec.linalg.givens_rotation(f, g)

    R = np.reshape([c, s, -s, c], (2,2))
    v = np.hstack([f, g])
    w = np.hstack([r, np.zeros_like(r)])
    res = R @ v - w
    np.testing.assert_allclose(res.astype(float), 0, atol=1e-31)


def test_givens():
    a = np.array([3.0, -2.0], dtype=ddouble)
    r, G = xprec.linalg.givens(a)
    diff = r - G @ a
    np.testing.assert_allclose(diff.astype(float), 0, atol=1e-31)


def test_qr():
    A = np.vander(np.linspace(-1, 1, 60), 80).astype(ddouble)
    Q, R = xprec.linalg.qr(A)
    I_m = np.eye(60)
    D = Q @ Q.T - I_m
    np.testing.assert_allclose(D.astype(float), 0, atol=4e-30)
    D = Q @ R - A
    np.testing.assert_allclose(D.astype(float), 0, atol=4e-30)


def test_qr_pivot():
    A = np.vander(np.linspace(-1, 1, 60), 80).astype(ddouble)
    Q, R, piv = xprec.linalg.rrqr(A)
    I_m = np.eye(60)
    D = Q @ Q.T - I_m
    np.testing.assert_allclose(D.astype(float), 0, atol=4e-30)

    D = Q @ R - A[:,piv]
    np.testing.assert_allclose(D.astype(float), 0, atol=4e-30)

    Rdiag = np.abs(R.diagonal())
    assert (Rdiag[1:] <= Rdiag[:-1]).all()


def test_jacobi():
    A = np.vander(np.linspace(-1, 1, 60), 80).astype(ddouble)
    U, s, VT = xprec.linalg.svd_trunc(A)
    np.testing.assert_allclose((U * s) @ VT - A, 0.0, atol=5e-30)
